#
# Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
from datetime import datetime
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import ARConfigGlobal, ARConfigIntf
from netq_lib.common import configuration
from netq_lib.common.enums import ServiceStatus, DefaultPaths, ConfKw, Platform, arConfigFiles
from netq_lib.common.utils import save_fullstate
from netq_lib.common.utils import check_cl_version_onwards, subprocess_wrapper, is_active_systemd_svc, get_logger
from netq_agent.parsers.common.render import BaseRender
from netq_agent.handlers.sdk_utils import SX_CHIP_TYPE_SPECTRUM2, SX_CHIP_TYPE_SPECTRUM3, SX_CHIP_TYPE_SPECTRUM4, SdkUtils

class AdaptiveRoutingConfigRender(BaseRender):
    def __init__(self):
        super(AdaptiveRoutingConfigRender, self).__init__()
        self.model = ARConfigGlobal
        self.last_config_modify_time = [None, None, None, None, None]
        self.is_cfg_in_asic = False
        self.fwd_state = False
        self.switchd_state_change_time = None
        self.__logger = get_logger(ConfKw.NETQ_AGENT, configuration)
        self.DEFAULT_LINK_UTIL_THRESHOLD = 70
        self.DEFAULT_PROFILE0_MODE = 0
        self.DEFAULT_CONG_THRESH_LOW = 200
        self.DEFAULT_CONG_THRESH_MED = 1000
        self.DEFAULT_CONG_THRESH_HIGH = 10000
        self.DEFAULT_SHAPER_TO = 10
        self.DEFAULT_SHAPER_FROM = 10
        self.DEFAULT_PROFILE0_BIND_TIME = 0
        self.DEFAULT_PROFILE0_FREE_THRESH = 4
        self.DEFAULT_PROFILE0_BUSY_THRESH = 0
        self.DEFAULT_PROFILE0_SHAPER_FROM_EN = True
        self.DEFAULT_PROFILE0_SHAPER_TO_EN = False
        self.DEFAULT_PROFILE0_ELEPHANT_FLOW = False
        self.DEFAULT_ECMP_SIZE = 64
        self.sdk_utils = None
        self.chip_type = None

    def is_ar_supported(self, platform):
        # ignore running these on VX
        if Platform.VX in platform:
            if os.path.exists(DefaultPaths.ARCONFIG_VX_FILE):
                # For testing alone, so scale AR config data for fabric without hardnodes
                return True
            return False

        # enable AR export only for Mellanox platform and on CL >= 5.1.0
        if not (check_cl_version_onwards(major=5, minor=1) and Platform.MELLANOX in platform):
            return False

        if Platform.CUMULUS not in platform:
            return False

        # enable AR export only for SPC2+
        if self.chip_type is None:
            self.sdk_utils = SdkUtils(self.__logger)
            self.chip_type = self.sdk_utils.get_chip_type_v3(platform)

        if self.chip_type and self.chip_type in (SX_CHIP_TYPE_SPECTRUM2, SX_CHIP_TYPE_SPECTRUM3, SX_CHIP_TYPE_SPECTRUM4):
            return True

        return False

    def is_cfg_in_asic_check(self, config_file_path, systemctl_timestamp):
        cfg_in_asic = False
        if os.path.exists(config_file_path):
            try:
                file_mtime = os.path.getmtime(config_file_path)

                # Convert the file modification time to a datetime object
                file_timestamp = datetime.fromtimestamp(file_mtime)
                if file_timestamp > systemctl_timestamp:
                    #switchd is not restarted/reloaded after modifying the config
                    cfg_in_asic = False
                else:
                    cfg_in_asic = True
            except Exception as e:
                self.__logger.info(e)
                pass
        else:
            cfg_in_asic = True
        return cfg_in_asic

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        ar_global_config = {}
        ar_intf_config = {}
        ar_profile_global_config = {}
        roce_enabled = False
        roce_mode = "None"
        fwd_state_changed = False
        fullstate_ar_global_config_arr = []
        fullstate_ar_intf_config_arr = []

        platform = kwargs.get('platform', '')
        if not self.is_ar_supported(platform):
            self.__logger.debug("AR: Unsupported platform {}".format(platform))
            return ServiceStatus.OK

        cmd_out = subprocess_wrapper('/bin/systemctl show -p StateChangeTimestamp switchd'.split())
        switchd_state_change_time = cmd_out.splitlines()[0].split('=')[1]

        is_switchd_active = is_active_systemd_svc('switchd.service')
        if is_switchd_active:
            if not self.fwd_state:
                # Old state was False, new is True. state change detected
                fwd_state_changed = True
            self.fwd_state = True
        else:
            if self.fwd_state:
                #Old state was True, new is False. state change detected
                fwd_state_changed = True
            self.fwd_state = False

        if self.switchd_state_change_time and self.switchd_state_change_time != switchd_state_change_time:
            fwd_state_changed = True
            self.switchd_state_change_time = switchd_state_change_time

        self.__logger.debug("AR: fwd_state {}, fwd_state_changed {}".format(self.fwd_state, fwd_state_changed))

        systemctl_timestamp = datetime.strptime(switchd_state_change_time, '%a %Y-%m-%d %H:%M:%S %Z')

        self.__logger.debug("AR: Config in Asic: {} fullstate: {}".format(self.is_cfg_in_asic, fullstate))

        if not fullstate and not self.is_ar_roce_config_changed():
            if self.is_cfg_in_asic and not fwd_state_changed:
                self.__logger.debug("AR: Config has not changed and is written to ASIC")
                # Config has not changed and is written to ASIC.
                # Nothing else to do
                return ServiceStatus.OK

        if os.path.exists(DefaultPaths.ROCE_STATUS_FILE_PATH):
            self.__logger.debug("AR: File exists {}".format(DefaultPaths.ROCE_STATUS_FILE_PATH))
            try:
                with open(DefaultPaths.ROCE_STATUS_FILE_PATH, 'r') as fd:
                    self.__logger.debug("AR: Open the file")
                    for line in fd.readlines():
                        self.__logger.debug("AR: line: {}".format(line))
                        if 'traffic.roce_mode' in line:
                            mode = int(line.strip().split('=')[1].strip())
                            if mode != 0:
                                roce_enabled = True

                            if mode == 1:
                                roce_mode = "Lossy"
                            elif mode == 3:
                                roce_mode = "Lossless"

                            self.__logger.debug("AR: roce_enable: {} roce_mode {}".format(roce_enabled, roce_mode))
                            break
            except Exception as e:
                self.__logger.info(e)
                pass

        is_roce_cfg_in_asic = self.is_cfg_in_asic_check(DefaultPaths.ROCE_STATUS_FILE_PATH, systemctl_timestamp)
        is_ar_cfg_in_asic = self.is_cfg_in_asic_check(DefaultPaths.ARCONFIG_FILE, systemctl_timestamp)

        m_arconfigintf_objs = ARConfigIntf.memmodel()

        old_arconfigintf_objs = []
        current_arconfigintf_objs = []
        if m_arconfigintf_objs:
            old_arconfigintf_objs = [old_obj for old_obj in m_arconfigintf_objs.query.filter(hostname=HOSTNAME)]

        if os.path.exists(DefaultPaths.ARCONFIG_FILE):
            self.__logger.debug("AR: File exists {}".format(DefaultPaths.ARCONFIG_FILE))
            try:
                with open(DefaultPaths.ARCONFIG_FILE, 'r') as fd:
                    self.__logger.debug("AR: opened the file")
                    for line in fd.readlines():
                        self.__logger.debug("AR: {}".format(line))
                        if line.startswith('#') or line.strip() == "":
                            #commented line or empty line, Ignore
                            continue

                        parts = line.split(' = ')
                        self.__logger.debug("AR: parts: {}".format(parts))
                        key = parts[0].strip()
                        value = parts[1].strip()

                        if value.isdigit():
                            value = int(value)
                        elif value.upper() == 'TRUE':
                            value = True
                        elif value.upper() == 'FALSE':
                            value = False

                        key0 = key.split('.')[0]
                        key1 = key.split('.')[1]
                        self.__logger.debug("AR: key: {} value: {} key0: {} key1 {}".format(key, value, key0, key1))

                        if key0 == 'interface' and 'sw' in key1:
                            intf = key1  # interface name
                            param = key.split('.')[3]  # Extracting the param

                            if intf not in ar_intf_config:
                                ar_intf_config[intf] = {}

                            ar_intf_config[intf][param] = value
                        else:
                            key = key.replace('adaptive_routing.', '')
                            ar_global_config[key] = value
            except Exception as e:
                self.__logger.info(e)
                pass

            # Accessing individual variables per port
            for ifname, ifparam in ar_intf_config.items():
                ar_intf_config_obj = ARConfigIntf()
                ar_intf_enable = ifparam.get('enable', False)
                ar_intf_link_util_thresh = ifparam.get('link_util_thresh', self.DEFAULT_LINK_UTIL_THRESHOLD)
                ar_intf_config_obj.ifname = ifname
                ar_intf_config_obj.is_enabled = ar_intf_enable
                ar_intf_config_obj.link_util_thresh = ar_intf_link_util_thresh
                current_arconfigintf_objs.append(ar_intf_config_obj)
                self.__logger.debug("AR: {}".format(ar_intf_config_obj))
                if not fullstate:
                    ar_intf_config_obj.save()
                    self.__logger.debug("AR: save interface config")
                else:
                    self.__logger.debug("AR: fullstate append interface config")
                    fullstate_ar_intf_config_arr.append(ar_intf_config_obj)

            if fullstate:
                save_fullstate(fullstate_ar_intf_config_arr, ARConfigIntf().__class__.__name__)

            if current_arconfigintf_objs:
                for old_obj in old_arconfigintf_objs:
                    if old_obj.ifname not in [new_obj.ifname for new_obj in current_arconfigintf_objs]:
                        old_ent = ARConfigIntf()
                        old_ent.copy(old_obj)
                        old_ent.delete()
                        self.__logger.debug("AR: delete old obj {}".format(old_obj))
                        old_arconfigintf_objs.remove(old_obj)
            else:
                self.__logger.debug("AR: Delete all AR intf")
                self.delete_ar_intf_config()

            # Global config parsing
            ar_enable = ar_global_config.get('enable', False)
            ar_profile_name = ar_global_config.get('profile', 'default')
            if ar_profile_name == 'default':
                self.__logger.debug("AR: default profile")
                is_ar_profile_cfg_in_asic = True
                ar_profile0_mode = ar_global_config.get('profile0.mode', self.DEFAULT_PROFILE0_MODE)
                ar_congestion_threshold_low = ar_global_config.get('congestion_threshold.low', self.DEFAULT_CONG_THRESH_LOW)
                ar_congestion_threshold_medium = ar_global_config.get('congestion_threshold.medium', self.DEFAULT_CONG_THRESH_MED)
                ar_congestion_threshold_high = ar_global_config.get('congestion_threshold.high', self.DEFAULT_CONG_THRESH_HIGH)
                ar_shaper_to = ar_global_config.get('shaper_to', self.DEFAULT_SHAPER_TO)
                ar_shaper_from = ar_global_config.get('shaper_from', self.DEFAULT_SHAPER_FROM)
                ar_profile0_bind_time = ar_global_config.get('profile0.bind_time', self.DEFAULT_PROFILE0_BIND_TIME)
                ar_profile0_free_threshold = ar_global_config.get('profile0.free_threshold', self.DEFAULT_PROFILE0_FREE_THRESH)
                ar_profile0_busy_threshold = ar_global_config.get('profile0.busy_threshold', self.DEFAULT_PROFILE0_BUSY_THRESH)
                ar_profile0_shaper_from_enable = ar_global_config.get('profile0.shaper_from_enable', self.DEFAULT_PROFILE0_SHAPER_FROM_EN)
                ar_profile0_shaper_to_enable = ar_global_config.get('profile0.shaper_to_enable', self.DEFAULT_PROFILE0_SHAPER_TO_EN)
                ar_profile0_elephant_flow = ar_global_config.get('profile0.elephant_flow', self.DEFAULT_PROFILE0_ELEPHANT_FLOW)
                ar_ecmp_size = ar_global_config.get('ecmp_size', self.DEFAULT_ECMP_SIZE)
            else:
                self.__logger.debug("AR: non-default profile {}".format(ar_profile_name))
                profile_file = DefaultPaths.ARCONFIG_PROFILE_FILE_PREFIX + ar_profile_name.replace("-", "_") + '.conf'
                is_ar_profile_cfg_in_asic = self.is_cfg_in_asic_check(profile_file, systemctl_timestamp)
                if os.path.exists(profile_file):
                    self.__logger.debug("AR: profile_file exists {}".format(profile_file))
                    try:
                        with open(profile_file, 'r') as fd:
                            for line in fd.readlines():
                                self.__logger.debug("AR: {}".format(line))

                                if line.startswith('#') or line.strip() == "":
                                    #commented line or empty line, Ignore
                                    continue

                                parts = line.split(' = ')
                                key = parts[0].strip()
                                value = parts[1].strip()

                                if value.isdigit():
                                    value = int(value)
                                elif value.upper() == 'TRUE':
                                    value = True
                                elif value.upper() == 'FALSE':
                                    value = False

                                key0 = key.split('.')[0]
                                key1 = key.split('.')[1]
                                self.__logger.debug("AR: key: {}, value: {}, key0: {}, key1: {}".format(key, value, key0, key1))

                                if key0 == 'interface' and 'sw' in key1:
                                    self.__logger.info("AR: Unexpected keys {} {} in {}".format(key0, key1, profile_file))
                                else:
                                    key = key.replace('adaptive_routing.', '')
                                    ar_profile_global_config[key] = value

                    except Exception as e:
                        self.__logger.info(e)
                        pass

                ar_profile0_mode = ar_profile_global_config.get('profile0.mode', self.DEFAULT_PROFILE0_MODE)
                ar_congestion_threshold_low = ar_profile_global_config.get('congestion_threshold.low', self.DEFAULT_CONG_THRESH_LOW)
                ar_congestion_threshold_medium = ar_profile_global_config.get('congestion_threshold.medium', self.DEFAULT_CONG_THRESH_MED)
                ar_congestion_threshold_high = ar_profile_global_config.get('congestion_threshold.high', self.DEFAULT_CONG_THRESH_HIGH)
                ar_shaper_to = ar_profile_global_config.get('shaper_to', self.DEFAULT_SHAPER_TO)
                ar_shaper_from = ar_profile_global_config.get('shaper_from', self.DEFAULT_SHAPER_FROM)
                ar_profile0_bind_time = ar_profile_global_config.get('profile0.bind_time', self.DEFAULT_PROFILE0_BIND_TIME)
                ar_profile0_free_threshold = ar_profile_global_config.get('profile0.free_threshold', self.DEFAULT_PROFILE0_FREE_THRESH)
                ar_profile0_busy_threshold = ar_profile_global_config.get('profile0.busy_threshold', self.DEFAULT_PROFILE0_BUSY_THRESH)
                ar_profile0_shaper_from_enable = ar_profile_global_config.get('profile0.shaper_from_enable', self.DEFAULT_PROFILE0_SHAPER_FROM_EN)
                ar_profile0_shaper_to_enable = ar_profile_global_config.get('profile0.shaper_to_enable', self.DEFAULT_PROFILE0_SHAPER_TO_EN)
                ar_profile0_elephant_flow = ar_profile_global_config.get('profile0.elephant_flow', self.DEFAULT_PROFILE0_ELEPHANT_FLOW)
                ar_ecmp_size = ar_profile_global_config.get('ecmp_size', self.DEFAULT_ECMP_SIZE)


            # Derive if all the config is written to ASIC:
            self.is_cfg_in_asic = is_roce_cfg_in_asic and is_ar_cfg_in_asic and is_ar_profile_cfg_in_asic and self.fwd_state
            self.__logger.debug("AR: is_cfg_in_asic: {}".format(self.is_cfg_in_asic))

            ar_global_config_obj = ARConfigGlobal()
            ar_global_config_obj.is_enabled = ar_enable
            ar_global_config_obj.is_roce_enabled = roce_enabled
            ar_global_config_obj.fwd_state = self.fwd_state
            ar_global_config_obj.is_cfg_in_asic = self.is_cfg_in_asic
            ar_global_config_obj.roce_mode = roce_mode
            ar_global_config_obj.profile_name = ar_profile_name
            ar_global_config_obj.profile0_mode = ar_profile0_mode
            ar_global_config_obj.congestion_threshold_low = ar_congestion_threshold_low
            ar_global_config_obj.congestion_threshold_medium = ar_congestion_threshold_medium
            ar_global_config_obj.congestion_threshold_high = ar_congestion_threshold_high
            ar_global_config_obj.shaper_to = ar_shaper_to
            ar_global_config_obj.shaper_from = ar_shaper_from
            ar_global_config_obj.profile0_bind_time = ar_profile0_bind_time
            ar_global_config_obj.profile0_free_threshold = ar_profile0_free_threshold
            ar_global_config_obj.profile0_busy_threshold = ar_profile0_busy_threshold
            ar_global_config_obj.profile0_shaper_from_enable = ar_profile0_shaper_from_enable
            ar_global_config_obj.profile0_shaper_to_enable = ar_profile0_shaper_to_enable
            ar_global_config_obj.profile0_elephant_flow = ar_profile0_elephant_flow
            ar_global_config_obj.ecmp_size = ar_ecmp_size
            self.__logger.debug("AR: ar_global_config_obj: {}".format(ar_global_config_obj))
            if not fullstate:
                ar_global_config_obj.save()
                self.__logger.debug("AR: ar_global_config_obj: save obj")
            else:
                fullstate_ar_global_config_arr.append(ar_global_config_obj)
                self.__logger.debug("AR: ar_global_config_obj: fullstate append obj")

            if fullstate:
                save_fullstate(fullstate_ar_global_config_arr, ARConfigGlobal().__class__.__name__)
                self.__logger.debug("AR: save_fullstate")
        else:
            self.delete_all_ar_config()

        return ServiceStatus.OK

    def delete_all_ar_config(self):
        self.delete_ar_global_config()
        self.delete_ar_intf_config()

    def delete_ar_global_config(self):
        m_ar_global_config = ARConfigGlobal.memmodel()
        if m_ar_global_config:
            for entry in m_ar_global_config.query.filter(hostname=HOSTNAME):
                self.__logger.debug("AR: Delete global entry {}".format(entry))
                old_ar_global_config = ARConfigGlobal()
                old_ar_global_config.copy(entry)
                old_ar_global_config.delete()

    def delete_ar_intf_config(self):
        m_ar_intf_config = ARConfigIntf.memmodel()
        if m_ar_intf_config:
            for entry in m_ar_intf_config.query.filter(hostname=HOSTNAME):
                self.__logger.debug("AR: Delete intf entry {}".format(entry))
                old_ar_intf_config = ARConfigIntf()
                old_ar_intf_config.copy(entry)
                old_ar_intf_config.delete()

    def is_ar_roce_config_changed(self):
        ar_conf_changed = self.is_config_changed(DefaultPaths.ARCONFIG_FILE, arConfigFiles.AR_CONF.value)
        ar_custom_profile_conf_changed = self.is_config_changed(DefaultPaths.ARCONFIG_CUSTOM_PROFILE_FILE, arConfigFiles.AR_CUSTOM_PROFILE_CONF.value)
        ar_profile1_profile_conf_changed = self.is_config_changed(DefaultPaths.ARCONFIG_PROFILE1_PROFILE_FILE, arConfigFiles.AR_PROFILE1_PROFILE_CONF.value)
        ar_profile2_profile_conf_changed = self.is_config_changed(DefaultPaths.ARCONFIG_PROFILE2_PROFILE_FILE, arConfigFiles.AR_PROFILE2_PROFILE_CONF.value)
        ar_roce_status_conf_changed = self.is_config_changed(DefaultPaths.ROCE_STATUS_FILE_PATH, arConfigFiles.AR_ROCE_CONF.value)

        config_changed = ar_conf_changed or ar_custom_profile_conf_changed or ar_profile1_profile_conf_changed or ar_roce_status_conf_changed or \
                ar_profile2_profile_conf_changed

        if config_changed:
            self.is_cfg_in_asic = False

        if not self.fwd_state:
            # ASIC is not running, mark is_cfg_in_asic as False
            self.is_cfg_in_asic = False

        return config_changed

    def is_config_changed(self, config_file_path, idx):
        config_changed = False
        if os.path.exists(config_file_path):
            try:
                current_modify_time = int(os.path.getmtime(config_file_path))
                if self.last_config_modify_time[idx] is None:
                    config_changed = True
                elif current_modify_time > self.last_config_modify_time[idx]:
                    config_changed = True

                if config_changed:
                    self.last_config_modify_time[idx] = current_modify_time
            except Exception as e:
                pass

        return config_changed
